#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "linked_list.h"
#include "input.h"

typedef struct FORCE_FIELD {
  int nat;
  int nat_repl;
  CRYSTAL *crystal;
  
  ATOM *atoms;
  REPLICA_ATOM *atoms_repl;
  
  LIST *bond_list;
  LIST *angle_list;
  LIST *torsion_list;
  LIST *inversion_list;
  LIST *nonbonded_list;
  
} FORCE_FIELD;


typedef struct CONNECTIVITY {
  int ncon; // number of atoms this atom is bonded to
  int partners[8]; // indeces of atoms this atom is connected to,
  // It is assumed that each atom can have at most 8 bonding partners
} CONNECTIVITY;

//----------- Force Field Terms -----------------------

// harmonic bond E = 1/2 Ke (R - Re)^2
typedef struct BOND {
  int I,J;
  double Re;   // equilibrium bond length
  double Ke;   // harmonic force constant
  // energy due to this bond
  double energy;
  // gradient on atom I, gradient on atom J is -gradI
  double gradI[3];
} BOND;

// harmonic cosine angle, E = 1/2 Ce (cos(theta) - cos(Te))^2
typedef struct ANGLE {
  int I,J,K;
  int linear;
  double Ce;  // force constant
  double Te;  // equilibrium angle
  double cosTe;  // cosine of equilibrium angle
  // energy due to angle bend
  double energy;
  // gradients on atoms I, J and K
  double gradI[3];
  double gradJ[3];
  double gradK[3];
} ANGLE;

// torsion potential E = 1/2 V_JK (1 - cos(n*(phi - phi^0_JK)))
typedef struct TORSION {
  int I,J,K,L;
  int ntors;  // number of torsions around J-K bond
  // 
  int n;
  double Phi0;
  double cosPhi0;  // cosine of phi0
  double Vbarrier;
  // energy due to this torsion
  double energy;
  // gradients on atoms involved
  double gradI[3];
  double gradJ[3];
  double gradK[3];
  double gradL[3];
} TORSION;

// spectroscipy inversion 
typedef struct INVERSION {
  int I,J,K,L;  // atom I is bonded to three other atoms J,K,L
  double Ps0;   // inversion angle
  double cosPs0;
  double Kinv;   // force constant
  int planar;   // 1 if Psi=0
  // energy due to this improper torsion
  double energy;
  // gradients on atoms involved
  double gradI[3];
  double gradJ[3];
  double gradK[3];
  double gradL[3];
} INVERSION;

// 12-6 Lennard Jones form   E = A R^(-12) - B R^(-6)
typedef struct VAN_DER_WAALS {
  int I,J;
  double A;   // A = D0 * R0^12
  double B;   // B = D0 * R0^6
  // energy due to this vdW interaction
  double energy;
  // gradient on atom I, gradient on atom J is -gradI
  double gradI[3];

} VAN_DER_WAALS;
//------------------------------------------------------


FORCE_FIELD *build_force_field(int nat, ATOM *atoms, CRYSTAL *crystal, int verbose)
{
  FORCE_FIELD *ff;
  int n1,n2,n3; // enumerate replicas of unit cell
  int i,j,k,l; // enumerate atoms, also used as general index
  int N;
  int I,J,K,L,tmp; // indeces of replica atoms
  ATOM *atI, *atJ;
  REPLICA_ATOM *atIr, *atJr, *atKr, *atLr;
  double xIJ, yIJ, zIJ, rIJ;
  CONNECTIVITY *connectivities;
  CONNECTIVITY *conI, *conJ, *conK;

  NODE *node;
  
  LIST *bond_list;
  BOND *bond;
  
  LIST *nonbonded_list;
  VAN_DER_WAALS *vdW;
  
  LIST *angle_list;
  ANGLE *angle;

  LIST *torsion_list;
  TORSION *torsion;
  int n, ntors;  // number of torsions
  int Is[12];
  int Ls[12];
  
  LIST *inversion_list;
  INVERSION *inversion;
  
  double Rij, Rij6, Dij;
  int continue_flag;
  
  REPLICA_ATOM *at_r;
  int nc = 1; // number of neigbhour cells
  
  ff = (FORCE_FIELD *) malloc(sizeof(FORCE_FIELD));
  ff->nat = nat;
  ff->nat_repl = nat*((int) pow(2*nc+1,3));
  ff->crystal = crystal;
  ff->atoms = atoms;
  ff->atoms_repl = (REPLICA_ATOM *) malloc(sizeof(REPLICA_ATOM) * ff->nat_repl);
  // create all replica atoms
  k = 0;
  for (n1=-nc; n1<=nc; n1++) {
    for (n2=-nc; n2<=nc; n2++) {
      for (n3=-nc; n3<=nc; n3++) {
	// enumerate atoms in central unit cell
	for (i=0; i<nat; i++) {
	  at_r = &(ff->atoms_repl[k]);
	  // at_r is a replica of atom i
	  at_r->index = i;
	  if ((n1 == 0) && (n2 == 0) && (n3 == 0)) {
	    at_r->central = 1;
	  } else {
	    at_r->central = 0;
	  }
	  // relative to the i-th atom in the central unit cell the replica
	  // is displaced by the vector [tx,ty,tz]
	  at_r->tx = n1*crystal->ax + n2*crystal->bx + n3*crystal->cx;
	  at_r->ty = n1*crystal->ay + n2*crystal->by + n3*crystal->cy;
	  at_r->tz = n1*crystal->az + n2*crystal->bz + n3*crystal->cz;
	  
	  k++;
	}

      }
    }
  }
  if (verbose > 0) {
    printf("number of atoms: %d\n", nat);
    printf("number of replica atoms: %d\n", k);
  }
  N = ff->nat_repl;
  // connectivities
  connectivities = (CONNECTIVITY *) malloc(sizeof(CONNECTIVITY) * N);
  // BONDS
  bond_list = list_new();
  for (I=0; I<N; I++) {
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    // initialized connectivity of atom I
    conI = &(connectivities[I]);
    conI->ncon = 0;
    
    for (J=0; J<N; J++) {
      if (I == J) continue;
      atJr = &(ff->atoms_repl[J]);
      atJ = &(ff->atoms[atJr->index]);
      // distance vector
      xIJ = atJ->x - atI->x + atIr->tx - atJr->tx;
      yIJ = atJ->y - atI->y + atIr->ty - atJr->ty;
      zIJ = atJ->z - atI->z + atIr->tz - atJr->tz;
      rIJ = sqrt(xIJ*xIJ + yIJ*yIJ + zIJ*zIJ);
      if (rIJ < 1.3*(atI->covR0 + atJ->covR0)) {
	// atoms I and J are bonded
	if (conI->ncon < 8) {
	  conI->partners[conI->ncon] = J;
	  conI->ncon++;
	} else {
	  continue;
	}
	// at least one atom should lie in the unit cell
	if (!(atIr->central | atJr->central)) continue;
	// add (I,J) bond to list if I<J
	if (I < J) {
	  bond = (BOND *) malloc(sizeof(BOND));
	  bond->I = I;
	  bond->J = J;
	  // see eqn. (6) in the Dreiding paper
	  bond->Re = atI->valR0 + atJ->valR0 - RE_DELTA;
	  bond->Ke = KE_SINGLE_BOND;
	  list_append(bond_list, (void *) bond);
	  //printf("BOND(%d-%d)\n", I+1,J+1);
	}
      }
    }
  }
  ff->bond_list = bond_list;
  if (verbose > 0) {
    printf("number of bonds: %d\n", (int) list_length(bond_list));
  }
  // ANGLES 
  angle_list = list_new();
  for (J=0; J<N; J++) {
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    conJ = &(connectivities[J]);
    for (i=0; i<conJ->ncon; i++) {
      I = conJ->partners[i];
      atIr = &(ff->atoms_repl[I]);
      for (k=i+1; k<conJ->ncon; k++) {
	K = conJ->partners[k];
	if (K<I) {
	  // swap atoms
	  tmp = I;
	  I = K;
	  K = tmp;
	}
	atKr = &(ff->atoms_repl[K]);
	// at least one atom should lie in the unit cell
	if (!(atIr->central | atJr->central | atKr->central)) continue;	
	angle = (ANGLE *) malloc(sizeof(ANGLE));
	angle->I = I;
	angle->J = J;
	angle->K = K;
	angle->Ce = atJ->Ce;
	angle->Te = atJ->Te;
	angle->cosTe = atJ->cosTe;
	angle->linear = atJ->linear;
	list_append(angle_list, (void *) angle);
	//printf("ANGLE(%d-%d-%d)\n", I+1,J+1,K+1);
      }
    }
  }
  ff->angle_list = angle_list;
  if (verbose > 0) {
    printf("number of angles: %d\n", (int) list_length(angle_list));
  }
  // TORSIONS
  torsion_list = list_new();
  for(node=ff->bond_list->head; node!=NULL; node=node->next) {
    bond = (BOND *) node->element;
    J = bond->I; K = bond->J;
    atJr = &(ff->atoms_repl[J]);
    atKr = &(ff->atoms_repl[K]);
    conJ = &(connectivities[J]);
    conK = &(connectivities[K]);
    // cound number of torsions around J-K bond and
    // find the neighbouring bond I-J and K-L
    ntors = 0;
    for(i=0;i<conJ->ncon;i++) {
      I = conJ->partners[i];
      if (I == K) continue;
      for(l=0;l<conK->ncon;l++) {
	L = conK->partners[l];
	if (L == J) continue;
	atIr = &(ff->atoms_repl[I]);
	atLr = &(ff->atoms_repl[L]);
	// at least one atom should lie in the unit cell
	if (!(atIr->central | atJr->central | atKr->central | atLr->central)) continue;	
	if (ntors < 12) {
	  Is[ntors] = I;
	  Ls[ntors] = L;
	  ntors++;
	}
      }
    }
    if (ntors > 0) {
      // add all torsions around J-K bond
      for(n=0; n<ntors; n++) {
	torsion = (TORSION *) malloc(sizeof(TORSION));
	torsion->ntors = ntors;
	I = Is[n]; L = Ls[n];
	torsion->I = I;
	torsion->J = J;
	torsion->K = K;
	torsion->L = L;
	// force field parameters
	// TODO: distinguish different cases (a)-(g) in the Dreiding paper
	torsion->n = 2;
	torsion->Phi0 = 0.0; // should depend on J-K bond
	torsion->cosPhi0 = 1.0; //  ""
	torsion->Vbarrier = VbarrierD; // ""
	// divide by number of torsions
	torsion->Vbarrier /= (double) torsion->ntors;
	list_append(torsion_list, (void *) torsion);
	//printf("TORSION(%d-%d-%d-%d)\n", I+1,J+1,K+1,L+1);
	//
      }
    }
  }
  ff->torsion_list = torsion_list;
  if (verbose > 0) {
    printf("number of torsions: %d\n", (int) list_length(torsion_list));
  }
  // INVERSIONS
  inversion_list = list_new();
  for (I=0; I<N; I++) {
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    conI = &(connectivities[I]);
    if (conI->ncon == 3) {
      // atom I is bonded to exactly three atoms
      J = conI->partners[0]; K = conI->partners[1]; L = conI->partners[2];
      atJr = &(ff->atoms_repl[J]);
      atKr = &(ff->atoms_repl[K]);
      atLr = &(ff->atoms_repl[L]);
      // at least one atom should lie in the unit cell
      if (!(atIr->central | atJr->central | atKr->central | atLr->central)) continue;	
      inversion = (INVERSION *) malloc(sizeof(INVERSION));
      inversion->I = I;
      inversion->J = J;
      inversion->K = K;
      inversion->L = L;
      // parameters
      inversion->Ps0 = PSI0_INVERSION;
      inversion->cosPs0 = COS_PSI0_INVERSION;
      inversion->Kinv = K_INVERSION;
      if (fabs(inversion->Ps0 - 0.0) < 1.0e-10) {
	inversion->planar = 1;
      } else {
	inversion->planar = 0;
      }
      list_append(inversion_list, (void *) inversion);
    }
  }
  ff->inversion_list = inversion_list;
  if (verbose > 0) {
    printf("number of inversions: %d\n", (int) list_length(inversion_list));
  }
  // NON-BONDING
  nonbonded_list = list_new();
  for (I=0; I<N; I++) {
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    conI = &(connectivities[I]);
    for (J=I+1; J<N; J++) {
      continue_flag = 0;
      // check if J is bonded to I
      for (i=0; i<conI->ncon; i++) {
	if (J == conI->partners[i]) {
	  continue_flag = 1;
	  break;
	}
      }
      if (continue_flag == 1) continue;
      // check that I and J are not bonded to a common atom
      conJ = &(connectivities[J]);
      for (i=0; i<conI->ncon; i++) {
	for (j=0; j<conJ->ncon; j++) {
	  if (conI->partners[i] == conJ->partners[j]) {
	    continue_flag = 1;
	    break;
	  }
	}
	if (continue_flag == 1) break;
      }
      if (continue_flag == 1) continue;      
      atJr = &(ff->atoms_repl[J]);
      atJ = &(ff->atoms[atJr->index]);
      // at least one atom should lie in the unit cell
      if (!(atIr->central | atJr->central)) continue;
      // distance vector
      xIJ = atJ->x - atI->x + atIr->tx - atJr->tx;
      yIJ = atJ->y - atI->y + atIr->ty - atJr->ty;
      zIJ = atJ->z - atI->z + atIr->tz - atJr->tz;
      rIJ = sqrt(xIJ*xIJ + yIJ*yIJ + zIJ*zIJ);
      if (rIJ > NONBONDING_CUTOFF) continue;
      //
      vdW = (VAN_DER_WAALS *) malloc(sizeof(VAN_DER_WAALS));
      vdW->I = I;
      vdW->J = J;
      Rij = sqrt(atI->vdWR0 * atJ->vdWR0);
      Dij = sqrt(atI->vdWD0 * atJ->vdWD0);
      Rij6 = pow(Rij,6);
      vdW->B = Dij * Rij6;
      vdW->A = Dij * Rij6*Rij6;

      list_append(nonbonded_list, (void *) vdW);
      //printf("VDW(%d,%d)\n", I+1,J+1);
    }
  }
  ff->nonbonded_list = nonbonded_list;
  if (verbose > 0) {
    printf("number of non-bonded interactions: %d\n", (int) list_length(nonbonded_list));
  }
  
  free(connectivities);
  return ff;
}

inline void cross_product(double *a, double *b, double *cr) {
  // compute the cross product cr = a x b
  cr[0] = a[1]*b[2] - a[2]*b[1];
  cr[1] = a[2]*b[0] - a[0]*b[2];
  cr[2] = a[0]*b[1] - a[1]*b[0];
}

inline double dot_product(double *a, double *b) {
  double dot;
  dot = a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
  return dot;
}

inline double norm(double *a) {
  double a2;
  a2 = a[0]*a[0] + a[1]*a[1] + a[2]*a[2];
  return sqrt(a2);
}

// The following formulae are taken from Guanghua Gao' PhD thesis
// "Large Scale Molecular Simulations with Application to Polymers and Nano-scale Materials

inline double angle_gradient(double *A, double *B,
			     double *gradI, double *gradJ, double *gradK) {
  double cosTh;
  double nA, nB;
  double iAB, iA2, iB2;
  int i;
  nA = norm(A);
  nB = norm(B);
  cosTh = dot_product(A,B)/(nA*nB);
  // gradients
  iAB = 1.0/(nA*nB);
  iA2 = 1.0/(nA*nA);
  iB2 = 1.0/(nB*nB);

  for(i=0;i<3;i++) {
    gradI[i] =  iAB * B[i]        - cosTh *  iA2 * A[i];
    gradJ[i] = -iAB * (A[i]+B[i]) + cosTh * (iA2 * A[i] + iB2 * B[i]);
    gradK[i] =  iAB * A[i]        - cosTh               * iB2 * B[i];
  }
  return cosTh;
}

inline double torsion_gradient(double *F, double *G, double *H,
			       double *gradI, double *gradJ, double *gradK, double *gradL) {
  double A[3], B[3];
  double nA, nB;
  double cosPh;
  double AxF[3], BxF[3], AxG[3], BxG[3], AxH[3], BxH[3];
  int i;
  double iAB, iA2, iB2;
  // A = F x G
  cross_product(F,G, A);
  // B = H x G
  cross_product(H,G, B);
  nA = norm(A);
  nB = norm(B);
  cosPh = dot_product(A,B)/(nA*nB);
  // gradients
  cross_product(A,F, AxF);
  cross_product(B,F, BxF);
  cross_product(A,G, AxG);
  cross_product(B,G, BxG);
  cross_product(A,H, AxH);
  cross_product(B,H, BxH);
  iAB = 1.0/(nA*nB);
  iA2 = 1.0/(nA*nA);
  iB2 = 1.0/(nB*nB);
  
  for(i=0;i<3;i++) {
    gradI[i] = - iAB * BxG[i]                      + cosPh * iA2 * AxG[i];
    gradJ[i] =   iAB * (-AxH[i] + BxG[i] - BxF[i]) - cosPh * (iA2 * (AxG[i]-AxF[i]) - iB2*BxH[i]);
    gradK[i] =   iAB * ( AxG[i] + AxH[i] + BxF[i]) - cosPh * (iA2 * AxF[i] + iB2*(BxG[i] + BxH[i]));
    gradL[i] =  -iAB * AxG[i]                      + cosPh * iB2 * BxG[i];
  }
  return cosPh;
}

inline double inversion_gradient(double *U, double *V, double *B,
				 double *gradI, double *gradJ, double *gradK, double *gradL) {
  double A[3];
  double nA, nB;
  double sinPs;
  double AxV[3], BxV[3], AxU[3], BxU[3];
  int i;
  double iAB, iA2, iB2;
  // A = U x V
  cross_product(U,V, A);
  nA = norm(A);
  nB = norm(B);
  sinPs = dot_product(A,B)/(nA*nB);
  // gradients
  cross_product(A,V, AxV);
  cross_product(B,V, BxV);
  cross_product(A,U, AxU);
  cross_product(B,U, BxU);
  iAB = 1.0/(nA*nB);
  iA2 = 1.0/(nA*nA);
  iB2 = 1.0/(nB*nB);
  
  for(i=0;i<3;i++) {
    gradI[i] =  iAB * (-A[i] + BxV[i] - BxU[i]) - sinPs * (iA2*(AxV[i] - AxU[i]) - iB2 * B[i]);
    gradJ[i] = -iAB * BxV[i]                    + sinPs * iA2 * AxV[i];
    gradK[i] =  iAB * BxU[i]                    - sinPs * iA2 * AxU[i];
    gradL[i] =  iAB * A[i]                      - sinPs                          * iB2 * B[i];
  }
  return sinPs;
}

void evaluate_force_field(FORCE_FIELD *ff) {
  int i;
  int I,J,K,L;
  NODE *node;
  BOND *bond;
  ANGLE *angle;
  TORSION *torsion;
  INVERSION *inversion;
  VAN_DER_WAALS *vdW;

  
  ATOM *atI, *atJ, *atK, *atL;
  REPLICA_ATOM *atIr, *atJr, *atKr, *atLr;
  double gradI[3], gradJ[3], gradK[3], gradL[3];
  double rJI[3], rJK[3], rKL[3];
  // for bond and vdW interactions
  double xIJ, yIJ, zIJ, nIJ, dR;
  double rinv, ri2, ri6, ri8, ri12, ri14, fac;
  // for angle interactions
  double cosTh;
  double dcos;
  // for torsion interactions
  double cosPh, x, x2, x3, x4, x5, x6, cosn, dcosn;
  // for inversion interactions
  double sinPs, cosPs, dcosPs;
  double rIJ[3], rIK[3], rIL[3];
  
  // evaluate bonded interaction
  for(node=ff->bond_list->head; node!=NULL; node=node->next) {
    bond = (BOND *) node->element;
    I = bond->I; J = bond->J;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    // distance vector
    xIJ = atJ->x - atI->x + atIr->tx - atJr->tx;
    yIJ = atJ->y - atI->y + atIr->ty - atJr->ty;
    zIJ = atJ->z - atI->z + atIr->tz - atJr->tz;
    nIJ = sqrt(xIJ*xIJ + yIJ*yIJ + zIJ*zIJ);

    dR = nIJ - bond->Re; 
    // energy
    bond->energy = 0.5*bond->Ke*dR*dR;
    // gradient
    fac = -bond->Ke*dR/nIJ;
    bond->gradI[0] = fac * xIJ;
    bond->gradI[1] = fac * yIJ;
    bond->gradI[2] = fac * zIJ;
  }
  //printf("bonds evaluated\n");
  // angles
  for(node=ff->angle_list->head; node!=NULL; node=node->next) {
    angle = (ANGLE *) node->element;
    angle->energy = 0.0;
    I = angle->I; J = angle->J; K = angle->K;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    atKr = &(ff->atoms_repl[K]);
    atK = &(ff->atoms[atKr->index]);

    // vector J-->I
    rJI[0] = atI->x - atJ->x + atJr->tx - atIr->tx;
    rJI[1] = atI->y - atJ->y + atJr->ty - atIr->ty;
    rJI[2] = atI->z - atJ->z + atJr->tz - atIr->tz;
    // vector J-->K
    rJK[0] = atK->x - atJ->x + atJr->tx - atKr->tx;
    rJK[1] = atK->y - atJ->y + atJr->ty - atKr->ty;
    rJK[2] = atK->z - atJ->z + atJr->tz - atKr->tz;

    cosTh = angle_gradient(rJI, rJK, gradI, gradJ, gradK);

    if (angle->linear == 1) {
      // E = Ce * (1 + cos(theta_IJK))
      fac = angle->Ce;
      angle->energy = fac * (1.0 + cosTh);
      for(i=0;i<3;i++) {
	angle->gradI[i] = fac * gradI[i];
	angle->gradJ[i] = fac * gradJ[i];
	angle->gradK[i] = fac * gradK[i];
      }
    } else {
      // E = 1/2 * Ce * [cos(theta_IJK) - cos(Te)]^2
      dcos = cosTh - angle->cosTe;
      angle->energy = 0.5 * angle->Ce * dcos*dcos;
      fac = angle->Ce * dcos;
      for(i=0;i<3;i++) {
	angle->gradI[i] = fac * gradI[i];
	angle->gradJ[i] = fac * gradJ[i];
	angle->gradK[i] = fac * gradK[i];
      }      
    }
  }
  //printf("angles evaluated\n");
  // torsions
  for(node=ff->torsion_list->head; node!=NULL; node=node->next) {
    torsion = (TORSION *) node->element;
    torsion->energy = 0.0;
    I = torsion->I; J = torsion->J; K = torsion->K; L = torsion->L;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    atKr = &(ff->atoms_repl[K]);
    atK = &(ff->atoms[atKr->index]);
    atLr = &(ff->atoms_repl[L]);
    atL = &(ff->atoms[atLr->index]);
    // vector J-->I
    rJI[0] = atI->x - atJ->x + atJr->tx - atIr->tx;
    rJI[1] = atI->y - atJ->y + atJr->ty - atIr->ty;
    rJI[2] = atI->z - atJ->z + atJr->tz - atIr->tz;
    // vector J-->K
    rJK[0] = atK->x - atJ->x + atJr->tx - atKr->tx;
    rJK[1] = atK->y - atJ->y + atJr->ty - atKr->ty;
    rJK[2] = atK->z - atJ->z + atJr->tz - atKr->tz;
    // vector K-->L
    rKL[0] = atL->x - atK->x + atKr->tx - atLr->tx;
    rKL[1] = atL->y - atK->y + atKr->ty - atLr->ty;
    rKL[2] = atL->z - atK->z + atKr->tz - atLr->tz;

    cosPh = torsion_gradient(rJI, rJK, rKL,
			     gradI, gradJ, gradK, gradL);
    // use Chebyshev polynomials to evaluate cos(n*phi)
    // cos(n*phi) = sum_m T_n,m * cos(phi)^m
    //
    x = cosPh;
    if (torsion->n == 2) {
      x2 = x*x;
      cosn = 2.0*x2-1.0;
      // derivative
      dcosn = 4.0*x;
    } else if (torsion->n == 3) {
      x2 = x*x;
      x3 = x2*x;
      cosn = 4.0*x3 - 3.0*x;
      dcosn = 12.0*x2 - 3.0;
    } else if (torsion->n == 6) {
      x2 = x*x;
      x3 = x2*x;
      x4 = x2*x2;
      x5 = x3*x2;
      x6 = x3*x3;
      cosn = 32.0*x6 - 48.0*x4 + 18.0*x2 - 1.0;
      dcosn = 192.0*x5 - 192.0*x3 + 36.0*x;
    } else {
      cosn = 0.0;
      dcosn = 0.0;
      printf("WARNING: periodicity of torsion should be n=2,3 or 6!");
    }
    // cos(n*(phi-phi0)) = cos(n*phi)*cos(n*phi0)   provided that phi0 = k*pi/n
    // E = V * [1 - cos(n*(phi-phi0))]
    torsion->energy = torsion->Vbarrier * (1.0 - torsion->cosPhi0 * cosn);
    // compute gradients of torsion energy
    fac = -torsion->Vbarrier * torsion->cosPhi0 * dcosn;
    for(i=0;i<3;i++) {
      torsion->gradI[i] = fac * gradI[i];
      torsion->gradJ[i] = fac * gradJ[i];
      torsion->gradK[i] = fac * gradK[i];
      torsion->gradL[i] = fac * gradL[i];
    }
  }
  //printf("torsions evaluated\n");
  // inversions
  for(node=ff->inversion_list->head; node!=NULL; node=node->next) {
    inversion = (INVERSION *) node->element;
    inversion->energy = 0.0;
    I = inversion->I; J = inversion->J; K = inversion->K; L = inversion->L;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    atKr = &(ff->atoms_repl[K]);
    atK = &(ff->atoms[atKr->index]);
    atLr = &(ff->atoms_repl[L]);
    atL = &(ff->atoms[atLr->index]);
    // vector I-->J
    rIJ[0] = atJ->x - atI->x + atIr->tx - atJr->tx;
    rIJ[1] = atJ->y - atI->y + atIr->ty - atJr->ty;
    rIJ[2] = atJ->z - atI->z + atIr->tz - atJr->tz;
    // vector I-->K
    rIK[0] = atK->x - atI->x + atIr->tx - atKr->tx;
    rIK[1] = atK->y - atI->y + atIr->ty - atKr->ty;
    rIK[2] = atK->z - atI->z + atIr->tz - atKr->tz;
    // vector I-->L
    rIL[0] = atL->x - atI->x + atIr->tx - atLr->tx;
    rIL[1] = atL->y - atI->y + atIr->ty - atLr->ty;
    rIL[2] = atL->z - atI->z + atIr->tz - atLr->tz;

    // angle between I-->L  and the plane I-J-K
    sinPs = inversion_gradient(rIJ, rIK,  rIL,
			       gradI, gradJ, gradK, gradL);
    cosPs = sqrt(1.0 - sinPs*sinPs);
    dcosPs = cosPs - inversion->cosPs0;
    inversion->energy += 1.0/6.0 * inversion->Kinv * dcosPs*dcosPs;
    // gradient
    fac = -1.0/3.0 * inversion->Kinv * dcosPs * sinPs/cosPs;
    for(i=0;i<3;i++) {
      inversion->gradI[i] = fac * gradI[i];
      inversion->gradJ[i] = fac * gradJ[i];
      inversion->gradK[i] = fac * gradK[i];
      inversion->gradL[i] = fac * gradL[i];
    }
    // angle between I-->J  and the plane I-K-L
    sinPs = inversion_gradient(rIK, rIL,  rIJ,
			       gradI, gradK, gradL, gradJ);
    cosPs = sqrt(1.0 - sinPs*sinPs);
    dcosPs = cosPs - inversion->cosPs0;
    inversion->energy += 1.0/6.0 * inversion->Kinv * dcosPs*dcosPs;
    // gradient
    fac = -1.0/3.0 * inversion->Kinv * dcosPs * sinPs/cosPs;
    for(i=0;i<3;i++) {
      inversion->gradI[i] += fac * gradI[i];
      inversion->gradJ[i] += fac * gradJ[i];
      inversion->gradK[i] += fac * gradK[i];
      inversion->gradL[i] += fac * gradL[i];
    }
    // angle between I-->K  and the plane I-L-J
    sinPs = inversion_gradient(rIL, rIJ,  rIK,
			       gradI, gradL, gradJ, gradK);
    cosPs = sqrt(1.0 - sinPs*sinPs);
    dcosPs = cosPs - inversion->cosPs0;
    inversion->energy += 1.0/6.0 * inversion->Kinv * dcosPs*dcosPs;
    // gradient
    fac = -1.0/3.0 * inversion->Kinv * dcosPs * sinPs/cosPs;
    for(i=0;i<3;i++) {
      inversion->gradI[i] += fac * gradI[i];
      inversion->gradJ[i] += fac * gradJ[i];
      inversion->gradK[i] += fac * gradK[i];
      inversion->gradL[i] += fac * gradL[i];
    }
    
    
  }
  //printf("inversions evaluated\n");  
  // evaluate non-bonding interaction
  for(node=ff->nonbonded_list->head; node!=NULL; node=node->next) {
    vdW = (VAN_DER_WAALS *) node->element;
    I = vdW->I; J = vdW->J;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    // distance vector
    xIJ = atJ->x - atI->x + atIr->tx - atJr->tx;
    yIJ = atJ->y - atI->y + atIr->ty - atJr->ty;
    zIJ = atJ->z - atI->z + atIr->tz - atJr->tz;
    nIJ = sqrt(xIJ*xIJ + yIJ*yIJ + zIJ*zIJ);

    rinv = 1.0/nIJ;
    ri2 = rinv*rinv;
    ri6 = pow(rinv,6);
    ri8 = ri6*ri2;
    ri12 = ri6*ri6;
    ri14 = ri8*ri6;
    // energy
    vdW->energy = vdW->A * ri12 - vdW->B*ri6;
    // gradient
    fac = 12.0 * vdW->A * ri14 - 6.0 * vdW->B * ri8;
    vdW->gradI[0] = fac * xIJ;
    vdW->gradI[1] = fac * yIJ;
    vdW->gradI[2] = fac * zIJ;
  }
  //printf("vdWs evaluated\n");  
}

double collect_force_field(FORCE_FIELD *ff)
{
  int i,n;
  int I,J,K,L;
  NODE *node;
  BOND *bond;
  ANGLE *angle;
  TORSION *torsion;
  INVERSION *inversion;
  VAN_DER_WAALS *vdW;
  
  ATOM *atI, *atJ, *atK, *atL;
  REPLICA_ATOM *atIr, *atJr, *atKr, *atLr;

  double energy = 0.0;
  // set all gradients to 0
  for(n=0;n<ff->nat;n++) {
    atI = &(ff->atoms[n]);
    for(i=0;i<3;i++) {
      atI->grad[i] = 0.0;
    }
  }

  // bonds
  for(node=ff->bond_list->head; node!=NULL; node=node->next) {
    bond = (BOND *) node->element;
    I = bond->I; J = bond->J;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    // add energies
    energy += bond->energy;
    // add gradients
    for(i=0;i<3;i++) {
      atI->grad[i] += bond->gradI[i];
      atJ->grad[i] -= bond->gradI[i];
    }
  }
  //  angles
  for(node=ff->angle_list->head; node!=NULL; node=node->next) {
    angle = (ANGLE *) node->element;
    I = angle->I; J = angle->J; K = angle->K;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    atKr = &(ff->atoms_repl[K]);
    atK = &(ff->atoms[atKr->index]);
    // add energies
    energy += angle->energy;
    // add gradients
    for(i=0;i<3;i++) {
      atI->grad[i] += angle->gradI[i];
      atJ->grad[i] += angle->gradJ[i];
      atK->grad[i] += angle->gradK[i];
    }
  }
  // torsions
  for(node=ff->torsion_list->head; node!=NULL; node=node->next) {
    torsion = (TORSION *) node->element;
    I = torsion->I; J = torsion->J; K = torsion->K; L = torsion->L;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    atKr = &(ff->atoms_repl[K]);
    atK = &(ff->atoms[atKr->index]);
    atLr = &(ff->atoms_repl[L]);
    atL = &(ff->atoms[atLr->index]);
    // add energies
    energy += torsion->energy;
    // add gradients
    for(i=0;i<3;i++) {
      atI->grad[i] += torsion->gradI[i];
      atJ->grad[i] += torsion->gradJ[i];
      atK->grad[i] += torsion->gradK[i];
      atL->grad[i] += torsion->gradL[i];
    }
  }
  // inversions
  for(node=ff->inversion_list->head; node!=NULL; node=node->next) {
    inversion = (INVERSION *) node->element;
    I = inversion->I; J = inversion->J; K = inversion->K; L = inversion->L;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    atKr = &(ff->atoms_repl[K]);
    atK = &(ff->atoms[atKr->index]);
    atLr = &(ff->atoms_repl[L]);
    atL = &(ff->atoms[atLr->index]);
    // add energies
    energy += inversion->energy;
    // add gradients
    for(i=0;i<3;i++) {
      atI->grad[i] += inversion->gradI[i];
      atJ->grad[i] += inversion->gradJ[i];
      atK->grad[i] += inversion->gradK[i];
      atL->grad[i] += inversion->gradL[i];
    }
  }
  // non-bonding interactions
  for(node=ff->nonbonded_list->head; node!=NULL; node=node->next) {
    vdW = (VAN_DER_WAALS *) node->element;
    I = vdW->I; J = vdW->J;
    atIr = &(ff->atoms_repl[I]);
    atI = &(ff->atoms[atIr->index]);
    atJr = &(ff->atoms_repl[J]);
    atJ = &(ff->atoms[atJr->index]);
    // add energies
    energy += vdW->energy;
    // add gradients
    for(i=0;i<3;i++) {
      atI->grad[i] += vdW->gradI[i];
      atJ->grad[i] -= vdW->gradI[i];
    }
  }
  return energy;
}

void delete_force_field(FORCE_FIELD *ff)
{
  free(ff->atoms);
  free(ff->atoms_repl);
  list_delete(ff->bond_list);
  list_delete(ff->angle_list);
  list_delete(ff->torsion_list);
  list_delete(ff->inversion_list);
  list_delete(ff->nonbonded_list);
  free(ff);
}

void set_force_field_coordinates(FORCE_FIELD *ff, int n, double x, double y, double z)
{
  ATOM *atom;
  atom = &(ff->atoms[n]);
  atom->x = x;
  atom->y = y;
  atom->z = z;
}

void get_force_field_gradient(FORCE_FIELD *ff, int n, double *gradx, double *grady, double *gradz)
{
  ATOM *atom;
  atom = &(ff->atoms[n]);
  *gradx = atom->grad[0];
  *grady = atom->grad[1];
  *gradz = atom->grad[2];
}

int get_number_of_atoms(FORCE_FIELD *ff)
{
  return ff->nat;
}
